@@ -12,12 +12,32 @@ namespace cv {
 
 namespace kinfu {
 
-TSDFVolume::TSDFVolume(float _voxelSize, Affine3f _pose, float _raycastStepFactor, float _truncDist,
+static inline v_float32x4 tsdfToFloat_INTR(const v_int32x4& num)
+{
+    v_float32x4 num128 = v_setall_f32(-1.f / 128.f);
+    return v_cvt_f32(num) * num128;
+}
+
+static inline TsdfType floatToTsdf(float num)
+{
+    //CV_Assert(-1 < num <= 1);
+    int8_t res = int8_t(num * (-128.f));
+    res = res ? res : (num < 0 ? 1 : -1);
+    return res;
+}
+
+static inline float tsdfToFloat(TsdfType num)
+{
+    return float(num) * (-1.f / 128.f);
+}
+
+TSDFVolume::TSDFVolume(float _voxelSize, Matx44f _pose, float _raycastStepFactor, float _truncDist,
                        int _maxWeight, Point3i _resolution, bool zFirstMemOrder)
     : Volume(_voxelSize, _pose, _raycastStepFactor),
       volResolution(_resolution),
-      maxWeight(_maxWeight)
+      maxWeight( WeightType(_maxWeight) )
 {
+    CV_Assert(_maxWeight < 255);
     // Unlike original code, this should work with any volume size
     // Not only when (x,y,z % 32) == 0
     volSize   = Point3f(volResolution) * voxelSize;
@@ -55,8 +75,8 @@ TSDFVolume::TSDFVolume(float _voxelSize, Affine3f _pose, float _raycastStepFacto
 }
 
 // dimension in voxels, size in meters
-TSDFVolumeCPU::TSDFVolumeCPU(float _voxelSize, cv::Affine3f _pose, float _raycastStepFactor,
-                             float _truncDist, int _maxWeight, Point3i _resolution,
+TSDFVolumeCPU::TSDFVolumeCPU(float _voxelSize, cv::Matx44f _pose, float _raycastStepFactor,
+                             float _truncDist, int _maxWeight, Vec3i _resolution,
                              bool zFirstMemOrder)
     : TSDFVolume(_voxelSize, _pose, _raycastStepFactor, _truncDist, _maxWeight, _resolution,
                  zFirstMemOrder)
@@ -74,7 +94,7 @@ void TSDFVolumeCPU::reset()
     volume.forEach<VecTsdfVoxel>([](VecTsdfVoxel& vv, const int* /* position */)
     {
         TsdfVoxel& v = reinterpret_cast<TsdfVoxel&>(vv);
-        v.tsdf = 0.0f; v.weight = 0;
+        v.tsdf = floatToTsdf(0.0f); v.weight = 0;
     });
 }
 
@@ -86,7 +106,7 @@ TsdfVoxel TSDFVolumeCPU::at(const cv::Vec3i& volumeIdx) const
         (volumeIdx[2] >= volResolution.z || volumeIdx[2] < 0))
     {
         TsdfVoxel dummy;
-        dummy.tsdf   = 1.0f;
+        dummy.tsdf   = floatToTsdf(1.0f);
         dummy.weight = 0;
         return dummy;
     }
@@ -150,7 +170,7 @@ static inline depthType bilinearDepth(const Depth& m, cv::Point2f pt)
             if(b10) return v10;
             if(b11) return v11;
         }
-        else if(nz == 2)
+        if(nz == 2)
         {
             if(b00 && b10) v01 = v00, v11 = v10;
             if(b01 && b11) v00 = v01, v10 = v11;
@@ -159,7 +179,7 @@ static inline depthType bilinearDepth(const Depth& m, cv::Point2f pt)
             if(b00 && b11) v01 = v10 = (v00 + v11)*0.5f;
             if(b01 && b10) v00 = v11 = (v01 + v10)*0.5f;
         }
-        else if(nz == 3)
+        if(nz == 3)
         {
             if(!b00) v00 = v10 + v01 - v11;
             if(!b01) v01 = v00 + v11 - v10;
@@ -175,17 +195,21 @@ static inline depthType bilinearDepth(const Depth& m, cv::Point2f pt)
 }
 #endif
 
+
+
 struct IntegrateInvoker : ParallelLoopBody
 {
-    IntegrateInvoker(TSDFVolumeCPU& _volume, const Depth& _depth, const Intr& intrinsics, const cv::Affine3f& cameraPose,
-                     float depthFactor) :
+    IntegrateInvoker(TSDFVolumeCPU& _volume, const Depth& _depth, const Intr& intrinsics,
+                    const cv::Matx44f& cameraPose, float depthFactor, Mat _pixNorms) :
         ParallelLoopBody(),
         volume(_volume),
         depth(_depth),
+        intr(intrinsics),
         proj(intrinsics.makeProjector()),
-        vol2cam(cameraPose.inv() * _volume.pose),
+        vol2cam(Affine3f(cameraPose.inv()) * _volume.pose),
         truncDistInv(1.f/_volume.truncDist),
-        dfac(1.f/depthFactor)
+        dfac(1.f/depthFactor),
+        pixNorms(_pixNorms)
     {
         volDataStart = volume.volume.ptr<TsdfVoxel>();
     }
@@ -307,22 +331,27 @@ struct IntegrateInvoker : ParallelLoopBody
                     }
 
                     // norm(camPixVec) produces double which is too slow
-                    float pixNorm = sqrt(v_reduce_sum(camPixVec*camPixVec));
+                    int _u = (int) projected.get0();
+                    int _v = (int) v_rotate_right<1>(projected).get0();
+                    if (!(_u >= 0 && _u < depth.cols && _v >= 0 && _v < depth.rows))
+                        continue;
+                    float pixNorm = pixNorms.at<float>(_v, _u);
+                    // float pixNorm = sqrt(v_reduce_sum(camPixVec*camPixVec));
                     // difference between distances of point and of surface to camera
-                    TsdfType sdf = pixNorm*(v*dfac - zCamSpace);
+                    float sdf = pixNorm*(v*dfac - zCamSpace);
                     // possible alternative is:
                     // kftype sdf = norm(camSpacePt)*(v*dfac/camSpacePt.z - 1);
                     if(sdf >= -volume.truncDist)
                     {
-                        TsdfType tsdf = fmin(1.f, sdf * truncDistInv);
+                        TsdfType tsdf = floatToTsdf(fmin(1.f, sdf * truncDistInv));
 
                         TsdfVoxel& voxel = volDataY[z*volume.volDims[2]];
-                        int& weight      = voxel.weight;
+                        WeightType& weight      = voxel.weight;
                         TsdfType& value  = voxel.tsdf;
 
                         // update TSDF
-                        value  = (value*weight+tsdf) / (weight + 1);
-                        weight = min(weight + 1, volume.maxWeight);
+                        value  = floatToTsdf((tsdfToFloat(value)*weight+ tsdfToFloat(tsdf)) / (weight + 1));
+                        weight = (weight + 1) < volume.maxWeight ? (weight + 1) : volume.maxWeight;
                     }
                 }
             }
@@ -338,7 +367,7 @@ struct IntegrateInvoker : ParallelLoopBody
             {
                 TsdfVoxel* volDataY = volDataX+y*volume.volDims[1];
                 // optimization of camSpace transformation (vector addition instead of matmul at each z)
-                Point3f basePt = vol2cam*(Point3f(x, y, 0)*volume.voxelSize);
+                Point3f basePt = vol2cam*(Point3f(float(x), float(y), 0.0f)*volume.voxelSize);
                 Point3f camSpacePt = basePt;
                 // zStep == vol2cam*(Point3f(x, y, 1)*voxelSize) - basePt;
                 // zStep == vol2cam*[Point3f(x, y, 1) - Point3f(x, y, 0)]*voxelSize
@@ -348,7 +377,7 @@ struct IntegrateInvoker : ParallelLoopBody
                 int startZ, endZ;
                 if(abs(zStep.z) > 1e-5)
                 {
-                    int baseZ = -basePt.z / zStep.z;
+                    int baseZ = int(-basePt.z / zStep.z);
                     if(zStep.z > 0)
                     {
                         startZ = baseZ;
@@ -375,11 +404,13 @@ struct IntegrateInvoker : ParallelLoopBody
                 }
                 startZ = max(0, startZ);
                 endZ   = min(volume.volResolution.z, endZ);
-                for(int z = startZ; z < endZ; z++)
+
+                for (int z = startZ; z < endZ; z++)
                 {
                     // optimization of the following:
                     //Point3f volPt = Point3f(x, y, z)*volume.voxelSize;
                     //Point3f camSpacePt = vol2cam * volPt;
+
                     camSpacePt += zStep;
                     if(camSpacePt.z <= 0)
                         continue;
@@ -388,27 +419,31 @@ struct IntegrateInvoker : ParallelLoopBody
                     Point2f projected = proj(camSpacePt, camPixVec);
 
                     depthType v = bilinearDepth(depth, projected);
-                    if(v == 0)
+                    if (v == 0) {
                         continue;
+                    }
+
+                    int _u = projected.x;
+                    int _v = projected.y;
+                    if (!(_u >= 0 && _u < depth.cols && _v >= 0 && _v < depth.rows))
+                        continue;
+                    float pixNorm = pixNorms.at<float>(_v, _u);
 
-                    // norm(camPixVec) produces double which is too slow
-                    float pixNorm = sqrt(camPixVec.dot(camPixVec));
                     // difference between distances of point and of surface to camera
-                    TsdfType sdf = pixNorm*(v*dfac - camSpacePt.z);
-
+                    float sdf = pixNorm*(v*dfac - camSpacePt.z);
                     // possible alternative is:
                     // kftype sdf = norm(camSpacePt)*(v*dfac/camSpacePt.z - 1);
                     if(sdf >= -volume.truncDist)
                     {
-                        TsdfType tsdf = fmin(1.f, sdf * truncDistInv);
+                        TsdfType tsdf = floatToTsdf(fmin(1.f, sdf * truncDistInv));
 
                         TsdfVoxel& voxel = volDataY[z*volume.volDims[2]];
-                        int& weight      = voxel.weight;
+                        WeightType& weight      = voxel.weight;
                         TsdfType& value  = voxel.tsdf;
 
                         // update TSDF
-                        value  = (value*weight+tsdf) / (weight + 1);
-                        weight = min(weight + 1, volume.maxWeight);
+                        value  = floatToTsdf((tsdfToFloat(value)*weight+ tsdfToFloat(tsdf)) / (weight + 1));
+                        weight = min(int(weight + 1), int(volume.maxWeight));
                     }
                 }
             }
@@ -418,35 +453,73 @@ struct IntegrateInvoker : ParallelLoopBody
 
     TSDFVolumeCPU& volume;
     const Depth& depth;
+    const Intr& intr;
     const Intr::Projector proj;
     const cv::Affine3f vol2cam;
     const float truncDistInv;
     const float dfac;
     TsdfVoxel* volDataStart;
+    Mat pixNorms;
 };
 
+static cv::Mat preCalculationPixNorm(Depth depth, const Intr& intrinsics)
+{
+    int height = depth.rows;
+    int widht = depth.cols;
+    Point2f fl(intrinsics.fx, intrinsics.fy);
+    Point2f pp(intrinsics.cx, intrinsics.cy);
+    Mat pixNorm (height, widht, CV_32F);
+    std::vector<float> x(widht);
+    std::vector<float> y(height);
+    for (int i = 0; i < widht; i++)
+        x[i] = (i - pp.x) / fl.x;
+    for (int i = 0; i < height; i++)
+        y[i] = (i - pp.y) / fl.y;
+
+    for (int i = 0; i < height; i++)
+    {
+        for (int j = 0; j < widht; j++)
+        {
+            pixNorm.at<float>(i, j) = sqrtf(x[j] * x[j] + y[i] * y[i] + 1.0f);
+        }
+    }
+    return pixNorm;
+}
+
 // use depth instead of distance (optimization)
-void TSDFVolumeCPU::integrate(InputArray _depth, float depthFactor, const cv::Affine3f& cameraPose,
+void TSDFVolumeCPU::integrate(InputArray _depth, float depthFactor, const cv::Matx44f& cameraPose,
                               const Intr& intrinsics)
 {
     CV_TRACE_FUNCTION();
 
     CV_Assert(_depth.type() == DEPTH_TYPE);
+    CV_Assert(!_depth.empty());
     Depth depth = _depth.getMat();
-    IntegrateInvoker ii(*this, depth, intrinsics, cameraPose, depthFactor);
+    if (!(frameParams[0] == depth.rows  && frameParams[1] == depth.cols &&
+        frameParams[2] == intrinsics.fx && frameParams[3] == intrinsics.fy &&
+        frameParams[4] == intrinsics.cx && frameParams[5] == intrinsics.cy))
+    {
+        frameParams[0] = (float)depth.rows; frameParams[1] = (float)depth.cols;
+        frameParams[2] = intrinsics.fx;     frameParams[3] = intrinsics.fy;
+        frameParams[4] = intrinsics.cx;     frameParams[5] = intrinsics.cy;
+
+        pixNorms = preCalculationPixNorm(depth, intrinsics);
+    }
+
+    IntegrateInvoker ii(*this, depth, intrinsics, cameraPose, depthFactor, pixNorms);
     Range range(0, volResolution.x);
     parallel_for_(range, ii);
 }
 
 #if USE_INTRINSICS
 // all coordinate checks should be done in inclosing cycle
-inline TsdfType TSDFVolumeCPU::interpolateVoxel(Point3f _p) const
+inline float TSDFVolumeCPU::interpolateVoxel(Point3f _p) const
 {
     v_float32x4 p(_p.x, _p.y, _p.z, 0);
     return interpolateVoxel(p);
 }
 
-inline TsdfType TSDFVolumeCPU::interpolateVoxel(const v_float32x4& p) const
+inline float TSDFVolumeCPU::interpolateVoxel(const v_float32x4& p) const
 {
     // tx, ty, tz = floor(p)
     v_int32x4 ip  = v_floor(p);
@@ -472,7 +545,8 @@ inline TsdfType TSDFVolumeCPU::interpolateVoxel(const v_float32x4& p) const
     for(int i = 0; i < 8; i++)
         vx[i] = volData[neighbourCoords[i] + coordBase].tsdf;
 
-    v_float32x4 v0246(vx[0], vx[2], vx[4], vx[6]), v1357(vx[1], vx[3], vx[5], vx[7]);
+    v_float32x4 v0246 = tsdfToFloat_INTR(v_int32x4(vx[0], vx[2], vx[4], vx[6]));
+    v_float32x4 v1357 = tsdfToFloat_INTR(v_int32x4(vx[1], vx[3], vx[5], vx[7]));
     v_float32x4 vxx = v0246 + v_setall_f32(tz)*(v1357 - v0246);
 
     v_float32x4 v00_10 = vxx;
@@ -486,7 +560,7 @@ inline TsdfType TSDFVolumeCPU::interpolateVoxel(const v_float32x4& p) const
     return v0 + tx*(v1 - v0);
 }
 #else
-inline TsdfType TSDFVolumeCPU::interpolateVoxel(Point3f p) const
+inline float TSDFVolumeCPU::interpolateVoxel(Point3f p) const
 {
     int xdim = volDims[0], ydim = volDims[1], zdim = volDims[2];
 
@@ -501,19 +575,20 @@ inline TsdfType TSDFVolumeCPU::interpolateVoxel(Point3f p) const
     int coordBase = ix*xdim + iy*ydim + iz*zdim;
     const TsdfVoxel* volData = volume.ptr<TsdfVoxel>();
 
-    TsdfType vx[8];
-    for(int i = 0; i < 8; i++)
-        vx[i] = volData[neighbourCoords[i] + coordBase].tsdf;
+    float vx[8];
+    for (int i = 0; i < 8; i++)
+        vx[i] = tsdfToFloat(volData[neighbourCoords[i] + coordBase].tsdf);
 
-    TsdfType v00 = vx[0] + tz*(vx[1] - vx[0]);
-    TsdfType v01 = vx[2] + tz*(vx[3] - vx[2]);
-    TsdfType v10 = vx[4] + tz*(vx[5] - vx[4]);
-    TsdfType v11 = vx[6] + tz*(vx[7] - vx[6]);
+    float v00 = vx[0] + tz*(vx[1] - vx[0]);
+    float v01 = vx[2] + tz*(vx[3] - vx[2]);
+    float v10 = vx[4] + tz*(vx[5] - vx[4]);
+    float v11 = vx[6] + tz*(vx[7] - vx[6]);
 
-    TsdfType v0 = v00 + ty*(v01 - v00);
-    TsdfType v1 = v10 + ty*(v11 - v10);
+    float v0 = v00 + ty*(v01 - v00);
+    float v1 = v10 + ty*(v11 - v10);
 
     return v0 + tx*(v1 - v0);
+
 }
 #endif
 
@@ -566,7 +641,8 @@ inline v_float32x4 TSDFVolumeCPU::getNormalVoxel(const v_float32x4& p) const
             vx[i] = volData[neighbourCoords[i] + coordBase + 1*dim].tsdf -
                     volData[neighbourCoords[i] + coordBase - 1*dim].tsdf;
 
-        v_float32x4 v0246(vx[0], vx[2], vx[4], vx[6]), v1357(vx[1], vx[3], vx[5], vx[7]);
+        v_float32x4 v0246 = tsdfToFloat_INTR(v_int32x4(vx[0], vx[2], vx[4], vx[6]));
+        v_float32x4 v1357 = tsdfToFloat_INTR(v_int32x4(vx[1], vx[3], vx[5], vx[7]));
         v_float32x4 vxx = v0246 + v_setall_f32(tz)*(v1357 - v0246);
 
         v_float32x4 v00_10 = vxx;
@@ -581,8 +657,9 @@ inline v_float32x4 TSDFVolumeCPU::getNormalVoxel(const v_float32x4& p) const
     }
 
     v_float32x4 n       = v_load_aligned(an);
-    v_float32x4 invNorm = v_invsqrt(v_setall_f32(v_reduce_sum(n*n)));
-    return n*invNorm;
+    v_float32x4 Norm = v_sqrt(v_setall_f32(v_reduce_sum(n*n)));
+
+    return Norm.get0() < 0.0001f ? nanv : n/Norm;
 }
 #else
 inline Point3f TSDFVolumeCPU::getNormalVoxel(Point3f p) const
@@ -611,29 +688,32 @@ inline Point3f TSDFVolumeCPU::getNormalVoxel(Point3f p) const
         const int dim = volDims[c];
         float& nv = an[c];
 
-        TsdfType vx[8];
+        float vx[8];
         for(int i = 0; i < 8; i++)
-            vx[i] = volData[neighbourCoords[i] + coordBase + 1*dim].tsdf -
-                    volData[neighbourCoords[i] + coordBase - 1*dim].tsdf;
+            vx[i] = tsdfToFloat(volData[neighbourCoords[i] + coordBase + 1*dim].tsdf -
+                                volData[neighbourCoords[i] + coordBase - 1*dim].tsdf);
 
-        TsdfType v00 = vx[0] + tz*(vx[1] - vx[0]);
-        TsdfType v01 = vx[2] + tz*(vx[3] - vx[2]);
-        TsdfType v10 = vx[4] + tz*(vx[5] - vx[4]);
-        TsdfType v11 = vx[6] + tz*(vx[7] - vx[6]);
+        float v00 = vx[0] + tz*(vx[1] - vx[0]);
+        float v01 = vx[2] + tz*(vx[3] - vx[2]);
+        float v10 = vx[4] + tz*(vx[5] - vx[4]);
+        float v11 = vx[6] + tz*(vx[7] - vx[6]);
 
-        TsdfType v0 = v00 + ty*(v01 - v00);
-        TsdfType v1 = v10 + ty*(v11 - v10);
+        float v0 = v00 + ty*(v01 - v00);
+        float v1 = v10 + ty*(v11 - v10);
 
         nv = v0 + tx*(v1 - v0);
     }
 
-    return normalize(an);
+    float nv = sqrt(an[0] * an[0] +
+                    an[1] * an[1] +
+                    an[2] * an[2]);
+    return nv < 0.0001f ? nan3 : an / nv;
 }
 #endif
 
 struct RaycastInvoker : ParallelLoopBody
 {
-    RaycastInvoker(Points& _points, Normals& _normals, const Affine3f& cameraPose,
+    RaycastInvoker(Points& _points, Normals& _normals, const Matx44f& cameraPose,
                   const Intr& intrinsics, const TSDFVolumeCPU& _volume) :
         ParallelLoopBody(),
         points(_points),
@@ -647,8 +727,8 @@ struct RaycastInvoker : ParallelLoopBody
                                         volume.voxelSize,
                                         volume.voxelSize)),
         boxMin(),
-        cam2vol(volume.pose.inv() * cameraPose),
-        vol2cam(cameraPose.inv() * volume.pose),
+        cam2vol(volume.pose.inv() * Affine3f(cameraPose)),
+        vol2cam(Affine3f(cameraPose.inv()) * volume.pose),
         reproj(intrinsics.makeReprojector())
     {  }
 
@@ -731,7 +811,7 @@ struct RaycastInvoker : ParallelLoopBody
                     int zdim            = volume.volDims[2];
                     v_float32x4 rayStep = dir * v_setall_f32(tstep);
                     v_float32x4 next    = (orig + dir * v_setall_f32(tmin));
-                    TsdfType f = volume.interpolateVoxel(next), fnext = f;
+                    float f = volume.interpolateVoxel(next), fnext = f;
 
                     //raymarch
                     int steps = 0;
@@ -745,7 +825,7 @@ struct RaycastInvoker : ParallelLoopBody
                         int iz = ip.get0();
                         int coord = ix*xdim + iy*ydim + iz*zdim;
 
-                        fnext = volume.volume.at<TsdfVoxel>(coord).tsdf;
+                        fnext = tsdfToFloat(volume.volume.at<TsdfVoxel>(coord).tsdf);
                         if(fnext != f)
                         {
                             fnext = volume.interpolateVoxel(next);
@@ -763,8 +843,8 @@ struct RaycastInvoker : ParallelLoopBody
                     if(f > 0.f && fnext < 0.f)
                     {
                         v_float32x4 tp = next - rayStep;
-                        TsdfType ft    = volume.interpolateVoxel(tp);
-                        TsdfType ftdt  = volume.interpolateVoxel(next);
+                        float ft    = volume.interpolateVoxel(tp);
+                        float ftdt  = volume.interpolateVoxel(next);
                         float ts = tmin + tstep*(steps - ft/(ftdt - ft));
 
                         // avoid division by zero
@@ -810,7 +890,7 @@ struct RaycastInvoker : ParallelLoopBody
 
                 Point3f orig = camTrans;
                 // direction through pixel in volume space
-                Point3f dir = normalize(Vec3f(camRot * reproj(Point3f(x, y, 1.f))));
+                Point3f dir = normalize(Vec3f(camRot * reproj(Point3f(float(x), float(y), 1.f))));
 
                 // compute intersection of ray with all six bbox planes
                 Vec3f rayinv(1.f/dir.x, 1.f/dir.y, 1.f/dir.z);
@@ -823,8 +903,10 @@ struct RaycastInvoker : ParallelLoopBody
 
                 // near clipping plane
                 const float clip = 0.f;
-                float tmin = max(max(max(minAx.x, minAx.y), max(minAx.x, minAx.z)), clip);
-                float tmax =     min(min(maxAx.x, maxAx.y), min(maxAx.x, maxAx.z));
+                //float tmin = max(max(max(minAx.x, minAx.y), max(minAx.x, minAx.z)), clip);
+                //float tmax =     min(min(maxAx.x, maxAx.y), min(maxAx.x, maxAx.z));
+                float tmin = max({ minAx.x, minAx.y, minAx.z, clip });
+                float tmax = min({ maxAx.x, maxAx.y, maxAx.z });
 
                 // precautions against getting coordinates out of bounds
                 tmin = tmin + tstep;
@@ -838,11 +920,11 @@ struct RaycastInvoker : ParallelLoopBody
 
                     Point3f rayStep = dir * tstep;
                     Point3f next = (orig + dir * tmin);
-                    TsdfType f = volume.interpolateVoxel(next), fnext = f;
+                    float f = volume.interpolateVoxel(next), fnext = f;
 
                     //raymarch
                     int steps = 0;
-                    int nSteps = floor((tmax - tmin)/tstep);
+                    int nSteps = int(floor((tmax - tmin)/tstep));
                     for(; steps < nSteps; steps++)
                     {
                         next += rayStep;
@@ -852,11 +934,10 @@ struct RaycastInvoker : ParallelLoopBody
                         int ix = cvRound(next.x);
                         int iy = cvRound(next.y);
                         int iz = cvRound(next.z);
-                        fnext = volume.volume.at<TsdfVoxel>(ix*xdim + iy*ydim + iz*zdim).tsdf;
+                        fnext = tsdfToFloat(volume.volume.at<TsdfVoxel>(ix*xdim + iy*ydim + iz*zdim).tsdf);
                         if(fnext != f)
                         {
                             fnext = volume.interpolateVoxel(next);
-
                             // when ray crosses a surface
                             if(std::signbit(f) != std::signbit(fnext))
                                 break;
@@ -864,14 +945,13 @@ struct RaycastInvoker : ParallelLoopBody
                             f = fnext;
                         }
                     }
-
                     // if ray penetrates a surface from outside
                     // linearly interpolate t between two f values
                     if(f > 0.f && fnext < 0.f)
                     {
                         Point3f tp    = next - rayStep;
-                        TsdfType ft   = volume.interpolateVoxel(tp);
-                        TsdfType ftdt = volume.interpolateVoxel(next);
+                        float ft   = volume.interpolateVoxel(tp);
+                        float ftdt = volume.interpolateVoxel(next);
                         // float t = tmin + steps*tstep;
                         // float ts = t - tstep*ft/(ftdt - ft);
                         float ts = tmin + tstep*(steps - ft/(ftdt - ft));
@@ -892,7 +972,6 @@ struct RaycastInvoker : ParallelLoopBody
                         }
                     }
                 }
-
                 ptsRow[x] = toPtype(point);
                 nrmRow[x] = toPtype(normal);
             }
@@ -915,7 +994,7 @@ struct RaycastInvoker : ParallelLoopBody
 };
 
 
-void TSDFVolumeCPU::raycast(const cv::Affine3f& cameraPose, const Intr& intrinsics, Size frameSize,
+void TSDFVolumeCPU::raycast(const cv::Matx44f& cameraPose, const Intr& intrinsics, Size frameSize,
                             cv::OutputArray _points, cv::OutputArray _normals) const
 {
     CV_TRACE_FUNCTION();
@@ -981,7 +1060,7 @@ struct FetchPointsNormalsInvoker : ParallelLoopBody
             const TsdfVoxel& voxeld = volDataStart[(x+shift.x)*vol.volDims[0] +
                                                    (y+shift.y)*vol.volDims[1] +
                                                    (z+shift.z)*vol.volDims[2]];
-            TsdfType vd = voxeld.tsdf;
+            float vd = tsdfToFloat(voxeld.tsdf);
 
             if(voxeld.weight != 0 && vd != 1.f)
             {
@@ -1018,7 +1097,7 @@ struct FetchPointsNormalsInvoker : ParallelLoopBody
                 for(int z = 0; z < vol.volResolution.z; z++)
                 {
                     const TsdfVoxel& voxel0 = volDataY[z*vol.volDims[2]];
-                    TsdfType v0             = voxel0.tsdf;
+                    float v0             = tsdfToFloat(voxel0.tsdf);
                     if(voxel0.weight != 0 && v0 != 1.f)
                     {
                         Point3f V(Point3f((float)x + 0.5f, (float)y + 0.5f, (float)z + 0.5f)*vol.voxelSize);
@@ -1076,34 +1155,10 @@ void TSDFVolumeCPU::fetchPointsNormals(OutputArray _points, OutputArray _normals
     }
 }
 
-struct PushNormals
-{
-    PushNormals(const TSDFVolumeCPU& _vol, Normals& _nrm) :
-        vol(_vol), normals(_nrm), invPose(vol.pose.inv())
-    { }
-    void operator ()(const ptype &pp, const int * position) const
-    {
-        Point3f p = fromPtype(pp);
-        Point3f n = nan3;
-        if(!isNaN(p))
-        {
-            Point3f voxPt = (invPose * p);
-            voxPt         = voxPt * vol.voxelSizeInv;
-            n             = vol.pose.rotation() * vol.getNormalVoxel(voxPt);
-        }
-        normals(position[0], position[1]) = toPtype(n);
-    }
-    const TSDFVolumeCPU& vol;
-    Normals& normals;
-
-    Affine3f invPose;
-};
-
-
 void TSDFVolumeCPU::fetchNormals(InputArray _points, OutputArray _normals) const
 {
     CV_TRACE_FUNCTION();
-
+    CV_Assert(!_points.empty());
     if(_normals.needed())
     {
         Points points = _points.getMat();
@@ -1112,18 +1167,33 @@ void TSDFVolumeCPU::fetchNormals(InputArray _points, OutputArray _normals) const
         _normals.createSameSize(_points, _points.type());
         Normals normals = _normals.getMat();
 
-        points.forEach(PushNormals(*this, normals));
+        const TSDFVolumeCPU& _vol = *this;
+        auto PushNormals = [&](const ptype& pp, const int* position)
+        {
+            const TSDFVolumeCPU& vol(_vol);
+            Affine3f invPose(vol.pose.inv());
+            Point3f p = fromPtype(pp);
+            Point3f n = nan3;
+            if (!isNaN(p))
+            {
+                Point3f voxPt = (invPose * p);
+                voxPt = voxPt * vol.voxelSizeInv;
+                n = vol.pose.rotation() * vol.getNormalVoxel(voxPt);
+            }
+            normals(position[0], position[1]) = toPtype(n);
+        };
+        points.forEach(PushNormals);
     }
 }
 
 ///////// GPU implementation /////////
 
 #ifdef HAVE_OPENCL
-TSDFVolumeGPU::TSDFVolumeGPU(float _voxelSize, cv::Affine3f _pose, float _raycastStepFactor, float _truncDist, int _maxWeight,
+TSDFVolumeGPU::TSDFVolumeGPU(float _voxelSize, cv::Matx44f _pose, float _raycastStepFactor, float _truncDist, int _maxWeight,
                              Point3i _resolution) :
     TSDFVolume(_voxelSize, _pose, _raycastStepFactor, _truncDist, _maxWeight, _resolution, false)
 {
-    volume = UMat(1, volResolution.x * volResolution.y * volResolution.z, CV_32FC2);
+    volume = UMat(1, volResolution.x * volResolution.y * volResolution.z, CV_8UC2);
 
     reset();
 }
@@ -1137,12 +1207,54 @@ void TSDFVolumeGPU::reset()
     volume.setTo(Scalar(0, 0));
 }
 
+static cv::UMat preCalculationPixNormGPU(int depth_rows, int depth_cols, Vec2f fxy, Vec2f cxy)
+{
+    Mat x(1, depth_cols, CV_32F);
+    Mat y(1, depth_rows, CV_32F);
+    Mat _pixNorm(1, depth_rows * depth_cols, CV_32F);
+
+    for (int i = 0; i < depth_cols; i++)
+        x.at<float>(0, i) = (i - cxy[0]) / fxy[0];
+    for (int i = 0; i < depth_rows; i++)
+        y.at<float>(0, i) = (i - cxy[1]) / fxy[1];
+
+    cv::String errorStr;
+    cv::String name = "preCalculationPixNorm";
+    ocl::ProgramSource source = ocl::rgbd::tsdf_oclsrc;
+    cv::String options = "-cl-mad-enable";
+    ocl::Kernel kk;
+    kk.create(name.c_str(), source, options, &errorStr);
+
+
+    if (kk.empty())
+        throw std::runtime_error("Failed to create kernel: " + errorStr);
+
+    AccessFlag af = ACCESS_READ;
+    UMat pixNorm = _pixNorm.getUMat(af);
+    UMat xx = x.getUMat(af);
+    UMat yy = y.getUMat(af);
+
+    kk.args(ocl::KernelArg::PtrReadWrite(pixNorm),
+        ocl::KernelArg::PtrReadOnly(xx),
+        ocl::KernelArg::PtrReadOnly(yy),
+        depth_cols);
+
+    size_t globalSize[2];
+    globalSize[0] = depth_rows;
+    globalSize[1] = depth_cols;
+
+    if (!kk.run(2, globalSize, NULL, true))
+        throw std::runtime_error("Failed to run kernel");
+
+    return pixNorm;
+}
 
 // use depth instead of distance (optimization)
 void TSDFVolumeGPU::integrate(InputArray _depth, float depthFactor,
-                              const cv::Affine3f& cameraPose, const Intr& intrinsics)
+                              const cv::Matx44f& cameraPose, const Intr& intrinsics)
 {
     CV_TRACE_FUNCTION();
+    CV_Assert(!_depth.empty());
 
     UMat depth = _depth.getUMat();
 
@@ -1156,10 +1268,20 @@ void TSDFVolumeGPU::integrate(InputArray _depth, float depthFactor,
     if(k.empty())
         throw std::runtime_error("Failed to create kernel: " + errorStr);
 
-    cv::Affine3f vol2cam(cameraPose.inv() * pose);
+    cv::Affine3f vol2cam(Affine3f(cameraPose.inv()) * pose);
     float dfac = 1.f/depthFactor;
     Vec4i volResGpu(volResolution.x, volResolution.y, volResolution.z);
     Vec2f fxy(intrinsics.fx, intrinsics.fy), cxy(intrinsics.cx, intrinsics.cy);
+    if (!(frameParams[0] == depth.rows && frameParams[1] == depth.cols &&
+        frameParams[2] == intrinsics.fx && frameParams[3] == intrinsics.fy &&
+        frameParams[4] == intrinsics.cx && frameParams[5] == intrinsics.cy))
+    {
+        frameParams[0] = (float)depth.rows; frameParams[1] = (float)depth.cols;
+        frameParams[2] = intrinsics.fx;     frameParams[3] = intrinsics.fy;
+        frameParams[4] = intrinsics.cx;     frameParams[5] = intrinsics.cy;
+
+        pixNorms = preCalculationPixNormGPU(depth.rows, depth.cols, fxy, cxy);
+    }
 
     // TODO: optimization possible
     // Use sampler for depth (mask needed)
@@ -1174,7 +1296,8 @@ void TSDFVolumeGPU::integrate(InputArray _depth, float depthFactor,
            cxy.val,
            dfac,
            truncDist,
-           maxWeight);
+           int(maxWeight),
+           ocl::KernelArg::PtrReadOnly(pixNorms));
 
     size_t globalSize[2];
     globalSize[0] = (size_t)volResolution.x;
@@ -1185,7 +1308,7 @@ void TSDFVolumeGPU::integrate(InputArray _depth, float depthFactor,
 }
 
 
-void TSDFVolumeGPU::raycast(const cv::Affine3f& cameraPose, const Intr& intrinsics, Size frameSize,
+void TSDFVolumeGPU::raycast(const cv::Matx44f& cameraPose, const Intr& intrinsics, Size frameSize,
                             cv::OutputArray _points, cv::OutputArray _normals) const
 {
     CV_TRACE_FUNCTION();
@@ -1209,8 +1332,8 @@ void TSDFVolumeGPU::raycast(const cv::Affine3f& cameraPose, const Intr& intrinsi
     UMat normals = _normals.getUMat();
 
     UMat vol2camGpu, cam2volGpu;
-    Affine3f vol2cam = cameraPose.inv() * pose;
-    Affine3f cam2vol = pose.inv() * cameraPose;
+    Affine3f vol2cam = Affine3f(cameraPose.inv()) * pose;
+    Affine3f cam2vol = pose.inv() * Affine3f(cameraPose);
     Mat(cam2vol.matrix).copyTo(cam2volGpu);
     Mat(vol2cam.matrix).copyTo(vol2camGpu);
     Intr::Reprojector r = intrinsics.makeReprojector();
@@ -1251,6 +1374,7 @@ void TSDFVolumeGPU::raycast(const cv::Affine3f& cameraPose, const Intr& intrinsi
 void TSDFVolumeGPU::fetchNormals(InputArray _points, OutputArray _normals) const
 {
     CV_TRACE_FUNCTION();
+    CV_Assert(!_points.empty());
 
     if(_normals.needed())
     {
@@ -1417,7 +1541,7 @@ void TSDFVolumeGPU::fetchPointsNormals(OutputArray points, OutputArray normals)
 
 #endif
 
-cv::Ptr<TSDFVolume> makeTSDFVolume(float _voxelSize, cv::Affine3f _pose, float _raycastStepFactor,
+cv::Ptr<TSDFVolume> makeTSDFVolume(float _voxelSize, cv::Matx44f _pose, float _raycastStepFactor,
                                    float _truncDist, int _maxWeight, Point3i _resolution)
 {
 #ifdef HAVE_OPENCL
